import React, { ChangeEvent, useState } from 'react'
import classNames from 'classnames'
import { ArrowContainer, Popover, PopoverPosition } from 'react-tiny-popover'
import { RoomInfo, VirtualBackgroundType } from '@zocks-communications/client-sdk-js'
import { useZocksStore } from '../../app'
import { JoinOptions, selectIsConnecting, useLocalMedia, useLocalVirtualBackgrounds } from '../../store'
import { Video } from '../Media'
import zocksStyle from './JoinPage.module.scss'
import { VirtualBackgroundSelector, VirtualBackgroundSelectorProps } from '../VirtualBackgroundSelector'
import { ErrorResponse } from '../common'
import {
  AudioControlButton,
  AudioControlButtonProps,
  VideoControlButton,
  VideoControlButtonProps,
} from '../RoomControls'
import { getClassName, getMergedClassNames } from '../common'
import { RoomJoinSelect } from './model'

export const JoinPage: React.FC<{
  joinAction: (joinOptions: JoinOptions) => void
  createRoomAction: () => Promise<RoomInfo>
  initialRoomJoinInputValue?: string
  initialRoomJoinSelectValue?: 'id' | 'name'
  joinText?: string
  newMeetingText?: string
  createNewMeetingText?: string
  newMeetingDisabled?: boolean
  roomJoinInputLabel?: string
  roomJoinSelectLabel?: string
  roomJoinInputDisabled?: boolean
  roomJoinInputPlaceholder?: string
  videoEnabledByDefault?: boolean
  audioEnabledByDefault?: boolean
  hideInputSelect?: boolean
  onNewRoomCreated?: (roomInfo: RoomInfo) => void
  dividerText?: string
  controlsHeaderElement?: string | JSX.Element
  disabledVideoElement?: string | JSX.Element
  audioControlButtonElement?: React.FC<AudioControlButtonProps>
  videoControlButtonElement?: React.FC<VideoControlButtonProps>
  virtualBackgroundSelectorElement?: React.FC<VirtualBackgroundSelectorProps>
  joinOptions?: Partial<JoinOptions>
  styles?: { [key: string]: string }
}> = ({
  joinAction,
  createRoomAction,
  initialRoomJoinInputValue,
  initialRoomJoinSelectValue = 'id',
  joinText = 'Join',
  newMeetingText = 'New Meeting',
  createNewMeetingText = 'Create New Meeting',
  newMeetingDisabled = false,
  roomJoinInputLabel = 'Room id',
  roomJoinSelectLabel = 'Room join select',
  roomJoinInputDisabled = false,
  roomJoinInputPlaceholder = 'Enter a room id or name',
  videoEnabledByDefault = false,
  audioEnabledByDefault = false,
  hideInputSelect = false,
  onNewRoomCreated = null,
  dividerText = null,
  controlsHeaderElement = null,
  disabledVideoElement = null,
  audioControlButtonElement = null,
  videoControlButtonElement = null,
  virtualBackgroundSelectorElement = null,
  joinOptions = {},
  styles,
}) => {
  const isConnecting = useZocksStore(selectIsConnecting)
  const [roomJoinInput, setRoomJoinInput] = useState<string | undefined>(initialRoomJoinInputValue)
  const [roomJoinInputSelectValue, setRoomJoinInputSelectValue] = useState<RoomJoinSelect>(initialRoomJoinSelectValue)
  const [hasError, setHasError] = useState<boolean>(false)
  const [error, setError] = useState<string>('')
  const [isPopoverOpen, setPopoverOpen] = useState<boolean>(false)

  const {
    currentAudioDeviceId,
    currentVideoDeviceId,
    localAudioStream,
    localVideoStream,
    videoStreamIsLoading,
    toggleLocalAudioStream,
    toggleLocalVideoStream,
    selectAudioDevice,
    selectVideoDevice,
  } = useLocalMedia(videoEnabledByDefault, audioEnabledByDefault)

  const audioDeviceSettings = localAudioStream?.mediaStreamTrack?.getSettings()
  const videoDeviceSettings = localVideoStream?.mediaStreamTrack?.getSettings()

  const {
    availableVirtualBackgrounds,
    currentVirtualBackground,
    virtualBackgroundStream,
    virtualBackgroundIsLoading,
    changeVirtualBackground,
    stopLocalVirtualBackground,
  } = useLocalVirtualBackgrounds()

  const audioEnabled = !!localAudioStream
  const videoEnabled = !!localVideoStream || !!virtualBackgroundStream

  const afterJoinCallback = () => {
    localAudioStream && toggleLocalAudioStream()
    localVideoStream && toggleLocalVideoStream()
    currentVirtualBackground && stopLocalVirtualBackground()
  }

  const joinErrorCallback = (e: ErrorResponse | string) => {
    setError(typeof e === 'string' ? e : e.errors.map((v) => v.message).join(', '))
    setPopoverOpen(true)
    setHasError(true)
  }

  const baseJoinOptions: JoinOptions = {
    roomId: roomJoinInputSelectValue === 'id' ? roomJoinInput : '',
    roomName: roomJoinInputSelectValue === 'name' ? roomJoinInput : '',
    autoCreate: false,
    audioDeviceSettings,
    videoDeviceSettings,
    virtualBackground: currentVirtualBackground,
    afterJoinCallback,
    joinErrorCallback,
    autoSubscribe: true,
    ...joinOptions,
  }

  const onRoomJoinInputChange = (evt: ChangeEvent<HTMLInputElement>) => {
    setHasError(false)
    setPopoverOpen(false)
    setRoomJoinInput(evt.target.value)
  }

  const onRoomJoinInputKeyUp = (evt: React.KeyboardEvent<HTMLInputElement>) => {
    if (roomJoinInput && evt.key === 'Enter') {
      joinAction(baseJoinOptions)
    }
  }

  const onRoomJoinSelectChange = (evt: ChangeEvent<HTMLSelectElement>) => {
    setRoomJoinInputSelectValue(evt.target.value as RoomJoinSelect)
  }

  const createNewMeeting = () => {
    createRoomAction().then((roomInfo: RoomInfo) => onNewRoomCreated && onNewRoomCreated(roomInfo))
  }

  const mergedStyles = getMergedClassNames(zocksStyle, styles)

  const joinErrorTooltip = (position: PopoverPosition | undefined, childRect: ClientRect, popoverRect: ClientRect) => (
    <div className={getClassName('join_page__room-id-input-popover', mergedStyles)}>
      <ArrowContainer
        position={position}
        childRect={childRect}
        popoverRect={popoverRect}
        arrowColor={'black'}
        arrowSize={10}
        className={getClassName('popover-arrow-container', mergedStyles)}
        arrowClassName='popover-arrow'
      >
        <div className={getClassName('join-page__room-id-input-popover-text', mergedStyles)}>{error}</div>
      </ArrowContainer>
    </div>
  )

  const changeVideoDevice = (videoDevice: MediaDeviceInfo) => {
    selectVideoDevice(videoDevice)
    if (currentVirtualBackground) {
      changeVirtualBackground(currentVirtualBackground, videoDevice.deviceId)
    }
  }

  const selectVirtualBackground = (virtualBackground: VirtualBackgroundType | null) => {
    changeVirtualBackground(virtualBackground, currentVideoDeviceId || undefined)
  }

  const toggleVideoStream = () => {
    toggleLocalVideoStream()
    if (videoEnabled) {
      stopLocalVirtualBackground()
    } else if (currentVirtualBackground) {
      changeVirtualBackground(currentVirtualBackground, currentVideoDeviceId || undefined)
    }
  }

  const AudioControlButtonElement = audioControlButtonElement ?? AudioControlButton
  const VideoControlButtonElement = videoControlButtonElement ?? VideoControlButton
  const VirtualBackgroundSelectorElement = virtualBackgroundSelectorElement ?? VirtualBackgroundSelector

  return (
    <div className={getClassName('join-page', mergedStyles)}>
      <div className={getClassName('join-page__local-streams', mergedStyles)}>
        {localVideoStream || virtualBackgroundStream ? (
          <div className={getClassName('join-page__video-container', mergedStyles)}>
            <div className={getClassName('join-page__video-preview', mergedStyles)}>
              <Video videoStream={virtualBackgroundStream || localVideoStream!} />
            </div>
          </div>
        ) : (
          <div className={getClassName('join-page__no-video-icon', mergedStyles)}>{disabledVideoElement}</div>
        )}
        <div className={getClassName('join-page__local-stream-buttons', mergedStyles)}>
          <AudioControlButtonElement
            audioIsOn={audioEnabled}
            currentDeviceId={currentAudioDeviceId}
            toggleAudio={toggleLocalAudioStream}
            onDeviceSelected={selectAudioDevice}
            styles={mergedStyles}
          />
          <VideoControlButtonElement
            videoIsOn={videoEnabled}
            toggleVideo={toggleVideoStream}
            onDeviceSelected={changeVideoDevice}
            videoIsLoading={videoStreamIsLoading}
            currentDeviceId={currentVideoDeviceId}
            availableVirtualBackgrounds={availableVirtualBackgrounds}
            selectVirtualBackground={selectVirtualBackground}
            currentVirtualBackground={currentVirtualBackground}
            virtualBackgroundIsLoading={virtualBackgroundIsLoading}
            styles={mergedStyles}
          />
          <VirtualBackgroundSelectorElement
            currentVirtualBackground={currentVirtualBackground}
            availableVirtualBackgrounds={availableVirtualBackgrounds}
            selectVirtualBackground={selectVirtualBackground}
            virtualBackgroundIsLoading={virtualBackgroundIsLoading}
            styles={mergedStyles}
          />
        </div>
      </div>
      <div className={getClassName('join-page__controls', mergedStyles)}>
        <div className={getClassName('join-page__join-room-container', mergedStyles)}>
          {controlsHeaderElement}
          <Popover
            isOpen={isPopoverOpen}
            content={({ position, childRect, popoverRect }) => joinErrorTooltip(position, childRect, popoverRect)}
            positions={['top']}
          >
            <div className={getClassName('join-page__controls__input-wrapper', mergedStyles)}>
              <input
                onMouseEnter={() => setPopoverOpen(hasError)}
                onMouseLeave={() => setPopoverOpen(false)}
                className={`${getClassName('join-page__controls__input', mergedStyles)}  ${classNames({
                  [getClassName('join-page__room-id-input--error', mergedStyles)]: hasError,
                })}`}
                type='text'
                defaultValue={initialRoomJoinInputValue ?? ''}
                onChange={onRoomJoinInputChange}
                onKeyUp={onRoomJoinInputKeyUp}
                placeholder={roomJoinInputPlaceholder}
                aria-label={roomJoinInputLabel}
                disabled={roomJoinInputDisabled}
              />
              {!hideInputSelect && (
                <select
                  aria-label={roomJoinSelectLabel}
                  onChange={onRoomJoinSelectChange}
                  className={getClassName('join-page__controls__select', mergedStyles)}
                  value={roomJoinInputSelectValue}
                >
                  <option value='name'>Name</option>
                  <option value='id'>Id</option>
                </select>
              )}
            </div>
          </Popover>
          <button
            className={`${getClassName('join-page__button', mergedStyles)} ${getClassName(
              'join-page__join-button',
              mergedStyles
            )}`}
            onClick={() => joinAction(baseJoinOptions)}
            disabled={!roomJoinInput || isConnecting || hasError}
          >
            {joinText}
          </button>
        </div>
        <div className={getClassName('join-page__controls-divider', mergedStyles)}>
          {dividerText ? <span>{dividerText}</span> : ''}
        </div>
        <div className={getClassName('join-page__new-meeting-container', mergedStyles)}>
          <button
            className={`${getClassName('join-page__button', mergedStyles)} ${getClassName(
              'join-page__new-meeting-button',
              mergedStyles
            )}`}
            onClick={() =>
              joinAction({
                ...baseJoinOptions,
                roomId: '',
                roomName: '',
                autoCreate: true,
              })
            }
            disabled={isConnecting || newMeetingDisabled}
          >
            {newMeetingText}
          </button>
          <button
            className={`${getClassName('join-page__button', mergedStyles)} ${getClassName(
              'join-page__create-new-meeting-button',
              mergedStyles
            )}`}
            onClick={createNewMeeting}
            disabled={isConnecting || newMeetingDisabled}
          >
            {createNewMeetingText}
          </button>
        </div>
      </div>
    </div>
  )
}
